#include <jni.h>
#include <SLES/OpenSLES.h>
#include <SLES/OpenSLES_Android.h>
#include <SLES/OpenSLES_AndroidConfiguration.h>
#include <math.h>
#define NULL 0

SLObjectItf mEngineObj;
SLEngineItf mEngine;
SLObjectItf mOutputMixObj;
SLObjectItf mPlayerObj;
SLPlayItf mPlayer;
SLBufferQueueItf mPlayerQueue;
short* mBuffer;
int mLength;
int started = 0;

int current_buffer;

void stop() {

	if (mPlayerObj != NULL) {
		(*mPlayerObj)->Destroy(mPlayerObj);
		mPlayerObj = NULL;
	}

	if (mOutputMixObj != NULL) {
		(*mOutputMixObj)->Destroy(mOutputMixObj);
		mOutputMixObj = NULL;
	}
	if (mEngineObj != NULL) {
		(*mEngineObj)->Destroy(mEngineObj);
		mEngineObj = NULL; mEngine = NULL;
	}
}

int initializationError(){
       stop();
       return -1;
}

int start() {
        //Log::info("Starting SoundService.");
        SLresult lRes;
        const SLuint32 lEngineMixIIDCount = 1;
        const SLInterfaceID lEngineMixIIDs[]={SL_IID_ENGINE};
        const SLboolean lEngineMixReqs[]={SL_BOOLEAN_TRUE};
        const SLuint32 lOutputMixIIDCount=0;
        const SLInterfaceID lOutputMixIIDs[]={};
        const SLboolean lOutputMixReqs[]={};
        lRes = slCreateEngine(&mEngineObj, 0, NULL,
        lEngineMixIIDCount, lEngineMixIIDs, lEngineMixReqs);

        if (lRes != SL_RESULT_SUCCESS)
            return initializationError();

        lRes=(*mEngineObj)->Realize(mEngineObj,SL_BOOLEAN_FALSE);
        if (lRes != SL_RESULT_SUCCESS)
            return initializationError();

        lRes=(*mEngineObj)->GetInterface(mEngineObj,
        SL_IID_ENGINE, &mEngine);

        if (lRes != SL_RESULT_SUCCESS)
            return initializationError();

        lRes=(*mEngine)->CreateOutputMix(mEngine,&mOutputMixObj,lOutputMixIIDCount,lOutputMixIIDs,lOutputMixReqs);
        lRes=(*mOutputMixObj)->Realize(mOutputMixObj,SL_BOOLEAN_FALSE);
        return 1;
}

//this callback hander is called every time a buffer finishes playing
void mPlayerCallback(SLAndroidSimpleBufferQueueItf bq, void *context){
	SLresult lRes;
	int current_sample = current_buffer * 240;
	if(current_sample < mLength){
		short playBuffer[240];
		for(int i=0 ; i < 240 ; i++){
			if(i+current_sample < mLength){
			  playBuffer[i] = mBuffer[i+current_sample];
			}else{
			  playBuffer[i] = 0;
			}
		}
		current_buffer++;
		lRes = (*mPlayerQueue)->Enqueue(mPlayerQueue, playBuffer,240);
	}
}

int startSound(){
	 SLresult lRes;
	 // Set-up sound audio source.
	 SLDataLocator_AndroidSimpleBufferQueue lDataLocatorIn;
	 lDataLocatorIn.locatorType = SL_DATALOCATOR_ANDROIDSIMPLEBUFFERQUEUE;
	 // At most one buffer in the queue.
	 lDataLocatorIn.numBuffers = 1;
	 SLDataFormat_PCM lDataFormat;
	 lDataFormat.formatType = SL_DATAFORMAT_PCM;
	 lDataFormat.numChannels = 1; // Mono sound.
	 lDataFormat.samplesPerSec = SL_SAMPLINGRATE_48;
	 lDataFormat.bitsPerSample = SL_PCMSAMPLEFORMAT_FIXED_16;
	 lDataFormat.containerSize = SL_PCMSAMPLEFORMAT_FIXED_16;
	 lDataFormat.channelMask = SL_SPEAKER_FRONT_CENTER;
	 lDataFormat.endianness = SL_BYTEORDER_LITTLEENDIAN;
	 SLDataSource lDataSource;
	 lDataSource.pLocator = &lDataLocatorIn;
	 lDataSource.pFormat = &lDataFormat;

	 SLDataLocator_OutputMix lDataLocatorOut;
	 lDataLocatorOut.locatorType = SL_DATALOCATOR_OUTPUTMIX;
	 lDataLocatorOut.outputMix = mOutputMixObj;
	 SLDataSink lDataSink;
	 lDataSink.pLocator = &lDataLocatorOut;
	 lDataSink.pFormat = NULL;

	 const SLuint32 lSoundPlayerIIDCount = 2;
	 const SLInterfaceID lSoundPlayerIIDs[] = { SL_IID_PLAY, SL_IID_BUFFERQUEUE };
	 const SLboolean lSoundPlayerReqs[] = { SL_BOOLEAN_TRUE, SL_BOOLEAN_TRUE };
	 lRes = (*mEngine)->CreateAudioPlayer(mEngine, &mPlayerObj,&lDataSource, &lDataSink, lSoundPlayerIIDCount,lSoundPlayerIIDs, lSoundPlayerReqs);
	 if (lRes != SL_RESULT_SUCCESS)
		 goto ERROR;
	  lRes = (*mPlayerObj)->Realize(mPlayerObj, SL_BOOLEAN_FALSE);
	  if (lRes != SL_RESULT_SUCCESS)
		  goto ERROR;
	  lRes = (*mPlayerObj)->GetInterface(mPlayerObj, SL_IID_PLAY,&mPlayer);
	  if (lRes != SL_RESULT_SUCCESS)
		  goto ERROR;
	  lRes = (*mPlayerObj)->GetInterface(mPlayerObj,SL_IID_BUFFERQUEUE, &mPlayerQueue);
	  if (lRes != SL_RESULT_SUCCESS)
		  goto ERROR;
	  //register callback on the playerqueue
	 // lRes = (*mPlayerQueue)->RegisterCallback(mPlayerQueue,mPlayerCallback,NULL);
	  //if (lRes != SL_RESULT_SUCCESS)
	 //		  goto ERROR;
	  lRes = (*mPlayer)->SetPlayState(mPlayer,SL_PLAYSTATE_PLAYING);
	  if (lRes != SL_RESULT_SUCCESS)
		   goto ERROR;
	  return 1;
	  ERROR:
	  	  //packt::Log::error("Error while starting SoundPlayer");
	  	  return -1;
}

void fillBufferQueue(){

	double sampleRate = 48000.0;
	double frequency = 440.0;
	double amplitude = 0.95;
	double seconds = 0.5;
	double twoPiF = 2 * 3.14159265 * frequency;
	int length = (int) (seconds * sampleRate);
	float buffer[length];
	for (int sample = 0; sample < length; sample++) {
	  double time = sample / sampleRate;
	  buffer[sample] = (float) amplitude * sin(twoPiF * time);
	}

	short shortBuffer[length];
	int bufferIndex = 0;
	for (int i = 0; i < length; i++) {
	  short x = (short) (buffer[i] * 32767.0);
	  shortBuffer[i]=x;
	}
	mBuffer = shortBuffer;
	mLength = length;
}

void playSound(){
	current_buffer = 0;
	SLresult lRes;

	lRes = (*mPlayerQueue)->Enqueue(mPlayerQueue, mBuffer,mLength);

	//queue the first buffer
	//int current_sample = current_buffer * 240;
	//if(current_sample < mLength){
	//	short playBuffer[240];
	//	for(int i=0 ; i < 240 ; i++){
	//		if(i+current_sample < mLength){
	//		  playBuffer[i] = mBuffer[i+current_sample];
	//		}else{
	//		  playBuffer[i] = 0;
	//		}
	//	}
	//	current_buffer++;

	//}
	//SLresult lRes;
	//SLuint32 lPlayerState;
	//(*mPlayerObj)->GetState(mPlayerObj, &lPlayerState);
	//if (lPlayerState == SL_OBJECT_STATE_REALIZED) {
	//	short* lBuffer = (short*) mBuffer;
	//	int lLength = mLength;
		// Removes any sound from the queue.
		//lRes = (*mPlayerQueue)->Clear(mPlayerQueue);
		//if (lRes != SL_RESULT_SUCCESS)
		//	goto ERROR;
		// Plays the new sound.

		// lRes = (*mPlayerQueue)->Enqueue(mPlayerQueue, lBuffer,lLength);

		 //if (lRes != SL_RESULT_SUCCESS)
		//	 goto ERROR;
	//}
}

JNIEXPORT jint JNICALL Java_be_tarsos_dsp_android_SoundPlayer_playSound(JNIEnv * env, jobject obj, jint arg){
	if(started==0){
		start();
		fillBufferQueue();
		startSound();
		started=1;
	}
	playSound();
}

JNIEXPORT jint JNICALL Java_be_tarsos_dsp_android_SoundPlayer_destroy(JNIEnv * env, jobject obj, jint arg){
	stop();
}


