package ugent.be.lowlatencytest;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import android.app.Activity;
import android.app.PendingIntent;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.PackageManager;
import android.hardware.usb.UsbDevice;
import android.hardware.usb.UsbManager;
import android.media.AudioManager;
import android.os.Bundle;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.TextView;
import android.widget.Toast;
import be.tarsos.dsp.android.SoundPlayer;

public class MainActivity extends Activity {

    private final static String TAG = "MainActivity";
    private final static boolean DEBUG = true;

    private static final int ARDUINO_USB_VENDOR_ID = 0x2341;
    private static final int ARDUINO_UNO_USB_PRODUCT_ID = 0x01;
    private static final int ARDUINO_MEGA_2560_USB_PRODUCT_ID = 0x10;
    private static final int ARDUINO_MEGA_2560_R3_USB_PRODUCT_ID = 0x42;
    private static final int ARDUINO_UNO_R3_USB_PRODUCT_ID = 0x43;
    private static final int ARDUINO_MEGA_2560_ADK_R3_USB_PRODUCT_ID = 0x44;
    private static final int ARDUINO_MEGA_2560_ADK_USB_PRODUCT_ID = 0x3F;

    private final SoundPlayer p = new SoundPlayer();
    CheckBox checkBox;
    TextView output; 
    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);
        checkBox = (CheckBox) findViewById(R.id.checkBox1);
        
        //this call initializes the audio engine.
        
        
        final Button button = (Button) findViewById(R.id.button);
        button.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {
            	p.playSound(checkBox.isChecked());
            }
        });
        
        PackageManager pm = getApplicationContext().getPackageManager();
        boolean claimsFeature =
        pm.hasSystemFeature(PackageManager.FEATURE_AUDIO_LOW_LATENCY);
        String low  = "   Supports low latency audio: " + claimsFeature;
        
        
        AudioManager am = (AudioManager) getSystemService(Context.AUDIO_SERVICE);
        String sampleRate = am.getProperty(AudioManager.PROPERTY_OUTPUT_SAMPLE_RATE);
        String framesPerBuffer =
        am.getProperty(AudioManager.PROPERTY_OUTPUT_FRAMES_PER_BUFFER);
        String samplerate =  "   Output sample rate: " + sampleRate;
        String bufferSize =  "   Audio frames per buffer: " + framesPerBuffer;
     
        output  = (TextView) findViewById(R.id.txtoutput);
        appendOutput("Device info:");
        appendOutput(low);
        appendOutput(samplerate);
        appendOutput(bufferSize);
        appendOutput("");

        IntentFilter filter = new IntentFilter();
        filter.addAction(ArduinoCommunicatorService.DATA_RECEIVED_INTENT);
        filter.addAction(UsbManager.ACTION_USB_DEVICE_DETACHED);

        registerReceiver(mReceiver, filter);

        findDevice();
    }
    
    public void appendOutput(final String data){
    	MainActivity.this.runOnUiThread(new Runnable() {
    	    public void run() {
    	    	output.setText(output.getText() + data + "\n");
    	    }
    	});
    }

    private void findDevice() {
        UsbManager usbManager = (UsbManager) getSystemService(Context.USB_SERVICE);
        UsbDevice usbDevice = null;
        HashMap<String, UsbDevice> usbDeviceList = usbManager.getDeviceList();
        if (DEBUG) Log.d(TAG, "length: " + usbDeviceList.size());
        Iterator<UsbDevice> deviceIterator = usbDeviceList.values().iterator();
        if (deviceIterator.hasNext()) {
            UsbDevice tempUsbDevice = deviceIterator.next();

            // Print device information. If you think your device should be able
            // to communicate with this app, add it to accepted products below.
            if (DEBUG) Log.d(TAG, "VendorId: " + tempUsbDevice.getVendorId());
            if (DEBUG) Log.d(TAG, "ProductId: " + tempUsbDevice.getProductId());
            if (DEBUG) Log.d(TAG, "DeviceName: " + tempUsbDevice.getDeviceName());
            if (DEBUG) Log.d(TAG, "DeviceId: " + tempUsbDevice.getDeviceId());
            if (DEBUG) Log.d(TAG, "DeviceClass: " + tempUsbDevice.getDeviceClass());
            if (DEBUG) Log.d(TAG, "DeviceSubclass: " + tempUsbDevice.getDeviceSubclass());
            if (DEBUG) Log.d(TAG, "InterfaceCount: " + tempUsbDevice.getInterfaceCount());
            if (DEBUG) Log.d(TAG, "DeviceProtocol: " + tempUsbDevice.getDeviceProtocol());

            if (tempUsbDevice.getVendorId() == ARDUINO_USB_VENDOR_ID) {
                if (DEBUG) Log.i(TAG, "Arduino device found!");

                switch (tempUsbDevice.getProductId()) {
                    case ARDUINO_UNO_USB_PRODUCT_ID:
                        Toast.makeText(getBaseContext(), "Arduino Uno " + getString(R.string.found), Toast.LENGTH_SHORT).show();
                        usbDevice = tempUsbDevice;
                        break;
                    case ARDUINO_MEGA_2560_USB_PRODUCT_ID:
                        Toast.makeText(getBaseContext(), "Arduino Mega 2560 " + getString(R.string.found), Toast.LENGTH_SHORT).show();
                        usbDevice = tempUsbDevice;
                        break;
                    case ARDUINO_MEGA_2560_R3_USB_PRODUCT_ID:
                        Toast.makeText(getBaseContext(), "Arduino Mega 2560 R3 " + getString(R.string.found), Toast.LENGTH_SHORT).show();
                        usbDevice = tempUsbDevice;
                        break;
                    case ARDUINO_UNO_R3_USB_PRODUCT_ID:
                        Toast.makeText(getBaseContext(), "Arduino Uno R3 " + getString(R.string.found), Toast.LENGTH_SHORT).show();
                        usbDevice = tempUsbDevice;
                        break;
                    case ARDUINO_MEGA_2560_ADK_R3_USB_PRODUCT_ID:
                        Toast.makeText(getBaseContext(), "Arduino Mega 2560 ADK R3 " + getString(R.string.found), Toast.LENGTH_SHORT).show();
                        usbDevice = tempUsbDevice;
                        break;
                    case ARDUINO_MEGA_2560_ADK_USB_PRODUCT_ID:
                        Toast.makeText(getBaseContext(), "Arduino Mega 2560 ADK " + getString(R.string.found), Toast.LENGTH_SHORT).show();
                        usbDevice = tempUsbDevice;
                        break;
                }
            }
        }

        if (usbDevice == null) {
            if (DEBUG) Log.i(TAG, "No device found!");
            Toast.makeText(getBaseContext(), getString(R.string.no_device_found), Toast.LENGTH_LONG).show();
        } else {
            if (DEBUG) Log.i(TAG, "Device found!");
            Intent startIntent = new Intent(getApplicationContext(), ArduinoCommunicatorService.class);
            PendingIntent pendingIntent = PendingIntent.getService(getApplicationContext(), 0, startIntent, 0);
            usbManager.requestPermission(usbDevice, pendingIntent);
        }
    }

    long prevTimeMillis = 0;
    private ArrayList<Long> delays = new ArrayList<Long>();
    BroadcastReceiver mReceiver = new BroadcastReceiver() {
        String prevData = "";
        private void handleTransferedData(Intent intent, boolean receiving) {

           final byte[] newTransferedData = intent.getByteArrayExtra(ArduinoCommunicatorService.DATA_EXTRA);
            StringBuilder hexStr = new StringBuilder();


            for (int i = 0; i < newTransferedData.length; i++) {
                if (Character.isLetterOrDigit(newTransferedData[i])) {
                    hexStr.append(new String(new byte[] {newTransferedData[i]}));
                }
            }

           String data =  hexStr.toString();
           Log.i(TAG, "data: " + newTransferedData.length + " \"" + data + "\"");
           
           if(data.length() >= 6 && !data.equals(prevData)) {
               prevData = data;
               if (data.startsWith("D")) {
            	   long delay = Long.parseLong(data.replace("D",""));
                   delays.add(delay);
                   String msg = "Measured delay: " + delay/1000 + " ms";
                   Log.i(TAG,msg);
                   appendOutput(msg);
               } else if (data.startsWith("A")) {
            	   p.playSound(checkBox.isChecked());
            	   long current = System.currentTimeMillis();
            	   long delta = current - prevTimeMillis - 1000;
            	   prevTimeMillis = current;
            	   
            	   long startmicros = Long.parseLong(data.replace("A",""));
            	   
            	   String msg = "Play audio at: " + startmicros/1000 + " ms drift: " + delta +"ms";
                   Log.i(TAG, msg);
                   appendOutput(msg);
               }
           }
        }

        @Override
        public void onReceive(Context context, Intent intent) {
            final String action = intent.getAction();
            if (ArduinoCommunicatorService.DATA_RECEIVED_INTENT.equals(action)) {
                handleTransferedData(intent, true);
            }else if (UsbManager.ACTION_USB_DEVICE_DETACHED.equals(action)) {
                long sum = 0;
                long max = -1;
                long min =10000000;
                for(int i = 0 ; i < delays.size() ; i++){
                    sum = sum + delays.get(i);
                    max=Math.max(delays.get(i),max);
                    min=Math.min(delays.get(i),min);
                }
                String msg = "Average delay: " + sum/delays.size()/1000 + "\nMin delay: " + min/1000 + "\nMax delay: " + max/1000;
                appendOutput(msg);
                appendOutput("");
            }
        }
    };



    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        // Inflate the menu; this adds items to the action bar if it is present.
        //getMenuInflater().inflate(R.menu.menu_main, menu);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        // Handle action bar item clicks here. The action bar will
        // automatically handle clicks on the Home/Up button, so long
        // as you specify a parent activity in AndroidManifest.xml.
        int id = item.getItemId();

        //noinspection SimplifiableIfStatement
        if (id == R.id.action_settings) {
            return true;
        }

        return super.onOptionsItemSelected(item);
    }

    protected void onDestroy() {
        super.onDestroy();
        unregisterReceiver(mReceiver);
        p.destroyAll();
    }
}
